package com.milaboratory.core.sequencing.motif;

import com.milaboratory.core.sequence.nucleotide.NucleotideSequence;

import java.util.BitSet;

/**
 * Implementation of simplest nucleotide sequence motif.
 *
 * @author Bolotin Dmitriy <bolotin.dmitriy@gmail.com>
 */
public final class NucleotideMotif {
    private final BitSet data;
    private final int size;

    /**
     * Inner private constructor for reverseComplement motif
     *
     * @param data
     */
    private NucleotideMotif(BitSet data, int size) {
        this.data = data;
        this.size = size;
    }

    /**
     * Creates motif from it's String representation.
     *
     * @param motif
     */
    public NucleotideMotif(String motif) {
        this.data = new BitSet(motif.length() << 2);
        for (int i = 0; i < motif.length(); ++i) {
            byte[] codes = NucleotideWildcards.getCodes(motif.charAt(i));
            if (codes == null)
                throw new RuntimeException("Illegal nucleotide/wildcard: \"" + motif.charAt(i) + "\".");
            for (byte b : codes)
                data.set(b + (i << 2));
        }
        this.size = motif.length();
    }

    /**
     * Size of motif.
     *
     * @return
     */
    public final int size() {
        return size;
    }

    /**
     * Return true if motif matches sub-sequence of nucleotide sequence from (from) inclusive to (from + size())
     * exclusive.
     *
     * @param sequence target sequence
     * @param from     offset to start matching
     * @return
     */
    public final boolean matches(NucleotideSequence sequence, int from) {
        if (from < 0 || from >= sequence.size())
            throw new IllegalArgumentException();
        if (sequence.size() < from + size())
            return false;
        for (int i = 0; i < size; ++i)
            if (!data.get(sequence.codeAt(from + i) + (i << 2)))
                return false;
        return true;
    }

    /**
     * Return number of mismatches between motif and sub-sequence of nucleotide
     * sequence from (from) inclusive to (from + size()) exclusive.
     *
     * @param sequence target sequence
     * @param from offset to start matching
     * @return
     */
    /*public final int mismatchCount(NucleotideSequence sequence, int from) {
        if (from < 0 || from >= sequence.size())
            throw new IllegalArgumentException();
        if (sequence.size() < from + size())
            return -1;
        for (int i = 0; i < size; ++i)
            if (!data.get(sequence.codeAt(from + i) + (i << 2)))
                return false;
        return true;
    }*/


    /**
     * Return reverse-complement motif.
     */
    public final NucleotideMotif reverseComplement() {
        BitSet newStore = new BitSet(data.size());
        for (int i = 0; i < data.size(); ++i)
            if (data.get(data.size() - 1 - i))
                newStore.set(i);
        return new NucleotideMotif(newStore, size);
    }

    /**
     * Try to find this motif in sub-sequence.
     *
     * @param sequence target sequence
     * @param from     left subsequence border; inclusive
     * @param to       right subsequence border; exclusive
     * @return coordinate of first match (coordinate of first nucleotide in the match) or -1 if no matches found
     */
    public final int findMatch(NucleotideSequence sequence, int from, int to) {
        if (from < 0 || from >= sequence.size() || to < 0
                || to > sequence.size()
                || from > to)
            throw new IllegalArgumentException();
        to -= size;
        for (int i = from; i < to; ++i)
            if (matches(sequence, i))
                return i;
        return -1;
    }

    /**
     * Try to find this motif in sequence.
     *
     * @param sequence target sequence
     * @return coordinate of first match (coordinate of first nucleotide in the match) or -1 if no matches found
     */
    public final int findMatch(NucleotideSequence sequence) {
        return findMatch(sequence, 0, sequence.size());
    }

    /**
     * Direct motif querying method. Returns true if nucleotide with specified code is allowed on the specified
     * position.
     *
     * @param position position in the motif
     * @param code     nucleotide code (1..3)
     * @return
     */
    public final boolean get(int position, byte code) {
        return data.get(code + (position << 2));
    }

    public String toString() {
        char[] chars = new char[size];
        int offset;
        for (int i = 0; i < size; ++i) {
            offset = i << 2;
            chars[i] = NucleotideWildcards.getSymbol(data.get(offset), data.get(offset + 1),
                    data.get(offset + 2), data.get(offset + 3));
        }
        return new String(chars);
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        NucleotideMotif that = (NucleotideMotif) o;

        if (!data.equals(that.data)) return false;

        return true;
    }

    @Override
    public int hashCode() {
        return data.hashCode();
    }
}
